@@ -2,6 +2,8 @@
 #
 # SPDX-License-Identifier: MIT
 
-# pylint: disable=unused-import
+from ..contexts.project import build_create_parser as build_parser
 
-from ..contexts.project import build_create_parser as build_parser
\ No newline at end of file
+__all__ = [
+    'build_parser',
+]
