@@ -2,25 +2,29 @@
 #
 # SPDX-License-Identifier: MIT
 
+from enum import Enum
 import argparse
 import json
 import logging as log
 import os
 import os.path as osp
 import shutil
-from enum import Enum
 
-from datumaro.components.cli_plugin import CliPlugin
+import numpy as np
+
 from datumaro.components.dataset_filter import DatasetItemEncoder
 from datumaro.components.extractor import AnnotationType
-from datumaro.components.operations import (DistanceComparator,
-    ExactComparator, compute_ann_statistics, compute_image_statistics, mean_std)
-from datumaro.components.project import \
-    PROJECT_DEFAULT_CONFIG as DEFAULT_CONFIG
+from datumaro.components.operations import (
+    DistanceComparator, ExactComparator, compute_ann_statistics,
+    compute_image_statistics,
+)
+from datumaro.components.project import PROJECT_DEFAULT_CONFIG as DEFAULT_CONFIG
 from datumaro.components.project import Environment, Project
+from datumaro.components.validator import TaskType
+from datumaro.util import error_rollback, on_error_do
+from datumaro.util.os_util import make_file_name
 
-from ...util import (CliException, MultilineFormatter, add_subparser,
-    make_file_name)
+from ...util import CliException, MultilineFormatter, add_subparser
 from ...util.project import generate_next_file_name, load_project
 from .diff import DiffVisualizer
 
@@ -344,7 +348,8 @@ def export_command(args):
 
     if args.filter:
         dataset = dataset.filter(args.filter, **filter_args)
-    dataset.export(format=args.format, save_dir=dst_dir, **extra_args)
+    converter = project.env.converters[args.format]
+    converter.convert(dataset, save_dir=dst_dir, **extra_args)
 
     log.info("Project exported to '%s' as '%s'" % (dst_dir, args.format))
 
@@ -506,8 +511,8 @@ def build_diff_parser(parser_ctor=argparse.ArgumentParser):
     parser.add_argument('-o', '--output-dir', dest='dst_dir', default=None,
         help="Directory to save comparison results (default: do not save)")
     parser.add_argument('-v', '--visualizer',
-        default=DiffVisualizer.DEFAULT_FORMAT,
-        choices=[f.name for f in DiffVisualizer.Format],
+        default=DiffVisualizer.DEFAULT_FORMAT.name,
+        choices=[f.name for f in DiffVisualizer.OutputFormat],
         help="Output format (default: %(default)s)")
     parser.add_argument('--iou-thresh', default=0.5, type=float,
         help="IoU match threshold for detections (default: %(default)s)")
@@ -521,6 +526,7 @@ def build_diff_parser(parser_ctor=argparse.ArgumentParser):
 
     return parser
 
+@error_rollback
 def diff_command(args):
     first_project = load_project(args.project_dir)
     second_project = load_project(args.other_project_dir)
@@ -540,17 +546,14 @@ def diff_command(args):
     dst_dir = osp.abspath(dst_dir)
     log.info("Saving diff to '%s'" % dst_dir)
 
-    dst_dir_existed = osp.exists(dst_dir)
-    try:
-        visualizer = DiffVisualizer(save_dir=dst_dir, comparator=comparator,
-            output_format=args.visualizer)
-        visualizer.save_dataset_diff(
+    if not osp.exists(dst_dir):
+        on_error_do(shutil.rmtree, dst_dir, ignore_errors=True)
+
+    with DiffVisualizer(save_dir=dst_dir, comparator=comparator,
+            output_format=args.visualizer) as visualizer:
+        visualizer.save(
             first_project.make_dataset(),
             second_project.make_dataset())
-    except BaseException:
-        if not dst_dir_existed and osp.isdir(dst_dir):
-            shutil.rmtree(dst_dir, ignore_errors=True)
-        raise
 
     return 0
 
@@ -611,7 +614,7 @@ def ediff_command(args):
         output["matches"] = matches
 
     output_file = generate_next_file_name('diff', ext='.json')
-    with open(output_file, 'w') as f:
+    with open(output_file, 'w', encoding='utf-8') as f:
         json.dump(output, f, indent=4, sort_keys=True)
 
     print("Found:")
@@ -715,7 +718,7 @@ def stats_command(args):
 
     dst_file = generate_next_file_name('statistics', ext='.json')
     log.info("Writing project statistics to '%s'" % dst_file)
-    with open(dst_file, 'w') as f:
+    with open(dst_file, 'w', encoding='utf-8') as f:
         json.dump(stats, f, indent=4, sort_keys=True)
 
 def build_info_parser(parser_ctor=argparse.ArgumentParser):
@@ -792,6 +795,76 @@ def info_command(args):
 
     return 0
 
+def build_validate_parser(parser_ctor=argparse.ArgumentParser):
+    def _parse_task_type(s):
+        try:
+            return TaskType[s.lower()].name
+        except:
+            raise argparse.ArgumentTypeError("Unknown task type %s. Expected "
+                "one of: %s" % (s, ', '.join(t.name for t in TaskType)))
+
+
+    parser = parser_ctor(help="Validate project",
+        description="""
+            Validates project based on specified task type and stores
+            results like statistics, reports and summary in JSON file.
+        """,
+        formatter_class=MultilineFormatter)
+
+    parser.add_argument('-t', '--task_type', type=_parse_task_type,
+        help="Task type for validation, one of %s" % \
+            ', '.join(t.name for t in TaskType))
+    parser.add_argument('-s', '--subset', dest='subset_name', default=None,
+        help="Subset to validate (default: whole dataset)")
+    parser.add_argument('-p', '--project', dest='project_dir', default='.',
+        help="Directory of the project to validate (default: current dir)")
+    parser.add_argument('extra_args', nargs=argparse.REMAINDER, default=None,
+        help="Optional arguments for validator (pass '-- -h' for help)")
+    parser.set_defaults(command=validate_command)
+
+    return parser
+
+def validate_command(args):
+    project = load_project(args.project_dir)
+    dst_file_name = f'report-{args.task_type}'
+
+    dataset = project.make_dataset()
+    if args.subset_name is not None:
+        dataset = dataset.get_subset(args.subset_name)
+        dst_file_name += f'-{args.subset_name}'
+
+    try:
+        validator_type = project.env.validators[args.task_type]
+    except KeyError:
+        raise CliException("Validator type '%s' is not found" % args.task_type)
+
+    extra_args = {}
+    if hasattr(validator_type, 'parse_cmdline'):
+        extra_args = validator_type.parse_cmdline(args.extra_args)
+
+    validator = validator_type(**extra_args)
+    report = validator.validate(dataset)
+
+    def numpy_encoder(obj):
+        if isinstance(obj, np.generic):
+            return obj.item()
+
+    def _make_serializable(d):
+        for key, val in list(d.items()):
+            # tuple key to str
+            if isinstance(key, tuple):
+                d[str(key)] = val
+                d.pop(key)
+            if isinstance(val, dict):
+                _make_serializable(val)
+
+    _make_serializable(report)
+
+    dst_file = generate_next_file_name(dst_file_name, ext='.json')
+    log.info("Writing project validation results to '%s'" % dst_file)
+    with open(dst_file, 'w', encoding='utf-8') as f:
+        json.dump(report, f, indent=4, sort_keys=True,
+                  default=numpy_encoder)
 
 def build_parser(parser_ctor=argparse.ArgumentParser):
     parser = parser_ctor(
@@ -815,5 +888,6 @@ def build_parser(parser_ctor=argparse.ArgumentParser):
     add_subparser(subparsers, 'transform', build_transform_parser)
     add_subparser(subparsers, 'info', build_info_parser)
     add_subparser(subparsers, 'stats', build_stats_parser)
+    add_subparser(subparsers, 'validate', build_validate_parser)
 
     return parser
