@@ -1,31 +1,32 @@
-
-# Copyright (C) 2019-2020 Intel Corporation
+# Copyright (C) 2019-2021 Intel Corporation
 #
 # SPDX-License-Identifier: MIT
 
-from enum import Enum
+from enum import Enum, auto
 from glob import iglob
-from typing import List, Dict
-import numpy as np
+from typing import Callable, Dict, Iterable, List, Optional, Tuple
+import os
 import os.path as osp
 
+from attr import attrib, attrs
 import attr
-from attr import attrs, attrib
+import numpy as np
 
+from datumaro.components.errors import DatasetNotFoundError
+from datumaro.util import is_method_redefined
+from datumaro.util.attrs_util import default_if_none, not_empty
 from datumaro.util.image import Image
-from datumaro.util.attrs_util import not_empty, default_if_none
 
 
-AnnotationType = Enum('AnnotationType',
-    [
-        'label',
-        'mask',
-        'points',
-        'polygon',
-        'polyline',
-        'bbox',
-        'caption',
-    ])
+class AnnotationType(Enum):
+    label = auto()
+    mask = auto()
+    points = auto()
+    polygon = auto()
+    polyline = auto()
+    bbox = auto()
+    caption = auto()
+    cuboid_3d = auto()
 
 _COORDINATE_ROUNDING_DIGITS = 2
 
@@ -68,7 +69,7 @@ class LabelCategories(Categories):
             iterable ([type]): This iterable object can be:
             1)simple str - will generate one Category with str as name
             2)list of str - will interpreted as list of Category names
-            3)list of positional argumetns - will generate Categories
+            3)list of positional arguments - will generate Categories
             with this arguments
 
 
@@ -98,6 +99,7 @@ class LabelCategories(Categories):
         self._indices = indices
 
     def add(self, name: str, parent: str = None, attributes: dict = None):
+        assert name
         assert name not in self._indices, name
 
         index = len(self.items)
@@ -114,6 +116,9 @@ class LabelCategories(Categories):
     def __getitem__(self, idx):
         return self.items[idx]
 
+    def __contains__(self, idx):
+        return 0 <= idx and idx < len(self.items)
+
     def __len__(self):
         return len(self.items)
 
@@ -127,6 +132,21 @@ class Label(Annotation):
 
 @attrs(eq=False)
 class MaskCategories(Categories):
+    @classmethod
+    def generate(cls, size=255, include_background=True):
+        """
+        Generates a color map with the specified size.
+
+        If include_background is True, the result will include the item
+            "0: (0, 0, 0)", which is typically used as a background color.
+        """
+        from datumaro.util.mask_tools import generate_colormap
+        colormap = generate_colormap(size + (not include_background))
+        if not include_background:
+            colormap.pop(0)
+            colormap = { k - 1: v for k, v in colormap.items() }
+        return cls(colormap)
+
     colormap = attrib(factory=dict, validator=default_if_none(dict))
     _inverse_colormap = attrib(default=None,
         validator=attr.validators.optional(dict))
@@ -139,6 +159,15 @@ class MaskCategories(Categories):
                 self._inverse_colormap = invert_colormap(self.colormap)
         return self._inverse_colormap
 
+    def __contains__(self, idx: int) -> bool:
+        return idx in self.colormap
+
+    def __getitem__(self, idx: int) -> Tuple[int, int, int]:
+        return self.colormap[idx]
+
+    def __len__(self):
+        return len(self.colormap)
+
     def __eq__(self, other):
         if not super().__eq__(other):
             return False
@@ -158,6 +187,10 @@ class Mask(Annotation):
         default=None, kw_only=True)
     z_order = attrib(default=0, validator=default_if_none(int), kw_only=True)
 
+    def __attrs_post_init__(self):
+        if isinstance(self._image, np.ndarray):
+            self._image = self._image.astype(bool)
+
     @property
     def image(self):
         if callable(self._image):
@@ -167,10 +200,12 @@ class Mask(Annotation):
     def as_class_mask(self, label_id=None):
         if label_id is None:
             label_id = self.label
-        return self.image * label_id
+        from datumaro.util.mask_tools import make_index_mask
+        return make_index_mask(self.image, label_id)
 
     def as_instance_mask(self, instance_id):
-        return self.image * instance_id
+        from datumaro.util.mask_tools import make_index_mask
+        return make_index_mask(self.image, instance_id)
 
     def get_area(self):
         return np.count_nonzero(self.image)
@@ -203,7 +238,7 @@ class RleMask(Mask):
     @staticmethod
     def _lazy_decode(rle):
         from pycocotools import mask as mask_utils
-        return lambda: mask_utils.decode(rle).astype(np.bool)
+        return lambda: mask_utils.decode(rle)
 
     def get_area(self):
         from pycocotools import mask as mask_utils
@@ -221,30 +256,57 @@ class RleMask(Mask):
 class CompiledMask:
     @staticmethod
     def from_instance_masks(instance_masks,
-            instance_ids=None, instance_labels=None):
-        from datumaro.util.mask_tools import merge_masks
+            instance_ids=None, instance_labels=None, dtype=None):
+        from datumaro.util.mask_tools import make_index_mask
 
-        if instance_ids is not None:
+        if instance_ids:
             assert len(instance_ids) == len(instance_masks)
         else:
             instance_ids = [None] * len(instance_masks)
 
-        if instance_labels is not None:
+        if instance_labels:
             assert len(instance_labels) == len(instance_masks)
         else:
             instance_labels = [None] * len(instance_masks)
 
-        instance_masks = sorted(
-            zip(instance_masks, instance_ids, instance_labels),
-            key=lambda m: m[0].z_order)
+        instance_masks = sorted(enumerate(instance_masks),
+            key=lambda m: m[1].z_order)
+        instance_masks = ((m.image, 1 + j,
+                instance_ids[i] if instance_ids[i] is not None else 1 + j,
+                instance_labels[i] if instance_labels[i] is not None else m.label
+            ) for j, (i, m) in enumerate(instance_masks))
 
-        instance_mask = [m.as_instance_mask(id if id is not None else 1 + idx)
-            for idx, (m, id, _) in enumerate(instance_masks)]
-        instance_mask = merge_masks(instance_mask)
+        # 1. Avoid memory explosion on materialization of all masks
+        # 2. Optimize materialization calls
+        it = iter(instance_masks)
 
-        cls_mask = [m.as_class_mask(c) for m, _, c in instance_masks]
-        cls_mask = merge_masks(cls_mask)
-        return __class__(class_mask=cls_mask, instance_mask=instance_mask)
+        instance_map = [0]
+        class_map = [0]
+
+        m, idx, instance_id, class_id = next(it)
+        if not class_id:
+            idx = 0
+        index_mask = make_index_mask(m, idx, dtype=dtype)
+        instance_map.append(instance_id)
+        class_map.append(class_id)
+
+        for m, idx, instance_id, class_id in it:
+            if not class_id:
+                idx = 0
+            index_mask = np.where(m, idx, index_mask)
+            instance_map.append(instance_id)
+            class_map.append(class_id)
+
+        if np.array_equal(instance_map, range(idx + 1)):
+            merged_instance_mask = index_mask
+        else:
+            merged_instance_mask = np.array(instance_map,
+                dtype=np.min_scalar_type(instance_map))[index_mask]
+        dtype_mask = dtype if dtype else np.min_scalar_type(class_map)
+        merged_class_mask = np.array(class_map, dtype=dtype_mask)[index_mask]
+
+        return __class__(class_mask=merged_class_mask,
+            instance_mask=merged_instance_mask)
 
     def __init__(self, class_mask=None, instance_mask=None):
         self._class_mask = class_mask
@@ -318,6 +380,66 @@ class PolyLine(_Shape):
     def get_area(self):
         return 0
 
+
+@attrs(init=False)
+class Cuboid3d(Annotation):
+    _type = AnnotationType.cuboid_3d
+    _points = attrib(type=list, default=None)
+    label = attrib(converter=attr.converters.optional(int),
+        default=None, kw_only=True)
+
+    @_points.validator
+    def _points_validator(self, attribute, points):
+        if points is None:
+            points = [0, 0, 0,  0, 0, 0,  1, 1, 1]
+        else:
+            assert len(points) == 3 + 3 + 3, points
+            points = [round(p, _COORDINATE_ROUNDING_DIGITS) for p in points]
+        self._points = points
+
+    def __init__(self, position, rotation=None, scale=None, **kwargs):
+        assert len(position) == 3, position
+        if not rotation:
+            rotation = [0] * 3
+        if not scale:
+            scale = [1] * 3
+        kwargs.pop('points', None)
+        self.__attrs_init__(points=[*position, *rotation, *scale], **kwargs)
+
+    @property
+    def position(self):
+        """[x, y, z]"""
+        return self._points[0:3]
+
+    @position.setter
+    def _set_poistion(self, value):
+        # TODO: fix the issue with separate coordinate rounding:
+        # self.position[0] = 12.345676
+        # - the number assigned won't be rounded.
+        self.position[:] = \
+            [round(p, _COORDINATE_ROUNDING_DIGITS) for p in value]
+
+    @property
+    def rotation(self):
+        """[rx, ry, rz]"""
+        return self._points[3:6]
+
+    @rotation.setter
+    def _set_rotation(self, value):
+        self.rotation[:] = \
+            [round(p, _COORDINATE_ROUNDING_DIGITS) for p in value]
+
+    @property
+    def scale(self):
+        """[sx, sy, sz]"""
+        return self._points[6:9]
+
+    @scale.setter
+    def _set_scale(self, value):
+        self.scale[:] = \
+            [round(p, _COORDINATE_ROUNDING_DIGITS) for p in value]
+
+
 @attrs
 class Polygon(_Shape):
     _type = AnnotationType.polygon
@@ -335,16 +457,13 @@ class Polygon(_Shape):
         area = mask_utils.area(rle)[0]
         return area
 
-@attrs
+@attrs(init=False)
 class Bbox(_Shape):
     _type = AnnotationType.bbox
 
-    # will be overridden by attrs, then will be overridden again by us
-    # attrs' method will be renamed to __attrs_init__
     def __init__(self, x, y, w, h, *args, **kwargs):
         kwargs.pop('points', None) # comes from wrap()
         self.__attrs_init__([x, y, x + w, y + h], *args, **kwargs)
-    __actual_init__ = __init__ # save pointer
 
     @property
     def x(self):
@@ -386,9 +505,6 @@ class Bbox(_Shape):
         d.update(kwargs)
         return attr.evolve(item, **d)
 
-assert not hasattr(Bbox, '__attrs_init__') # hopefully, it will be supported
-setattr(Bbox, '__attrs_init__', Bbox.__init__)
-setattr(Bbox, '__init__', Bbox.__actual_init__)
 
 @attrs
 class PointsCategories(Categories):
@@ -405,7 +521,7 @@ class PointsCategories(Categories):
 
         Args:
             iterable ([type]): This iterable object can be:
-            1) list of positional argumetns - will generate Categories
+            1) list of positional arguments - will generate Categories
                 with these arguments
 
         Returns:
@@ -423,13 +539,23 @@ class PointsCategories(Categories):
         joints = set(map(tuple, joints))
         self.items[label_id] = self.Category(labels, joints)
 
+    def __contains__(self, idx: int) -> bool:
+        return idx in self.items
+
+    def __getitem__(self, idx: int) -> Tuple[int, int, int]:
+        return self.items[idx]
+
+    def __len__(self):
+        return len(self.items)
+
+
 @attrs
 class Points(_Shape):
-    Visibility = Enum('Visibility', [
-        ('absent', 0),
-        ('hidden', 1),
-        ('visible', 2),
-    ])
+    class Visibility(Enum):
+        absent = 0
+        hidden = 1
+        visible = 2
+
     _type = AnnotationType.points
 
     visibility = attrib(type=list, default=None)
@@ -475,18 +601,42 @@ class DatasetItem:
     id = attrib(converter=lambda x: str(x).replace('\\', '/'),
         type=str, validator=not_empty)
     annotations = attrib(factory=list, validator=default_if_none(list))
-    subset = attrib(converter=lambda v: v or DEFAULT_SUBSET_NAME, default=None)
+    subset = attrib(converter=lambda v: v or DEFAULT_SUBSET_NAME,
+        type=str, default=None)
+
+    # Currently unused
     path = attrib(factory=list, validator=default_if_none(list))
 
-    image = attrib(type=Image, default=None)
-    @image.validator
-    def _image_validator(self, attribute, image):
+    # TODO: introduce "media" field with type info. Replace image and pcd.
+    image = attrib(type=Optional[Image], default=None)
+    # TODO: introduce pcd type like Image
+    point_cloud = attrib(converter=lambda x: \
+            str(x).replace('\\', '/') if x else None,
+        type=Optional[str], default=None)
+    related_images = attrib(type=List[Image], default=None)
+
+    def __attrs_post_init__(self):
+        if (self.has_image and self.has_point_cloud):
+            raise ValueError("Can't set both image and point cloud info")
+        if self.related_images and not self.has_point_cloud:
+            raise ValueError("Related images require point cloud")
+
+    def _image_converter(image):
         if callable(image) or isinstance(image, np.ndarray):
             image = Image(data=image)
         elif isinstance(image, str):
             image = Image(path=image)
-        assert image is None or isinstance(image, Image)
-        self.image = image
+        assert image is None or isinstance(image, Image), type(image)
+        return image
+    image.converter = _image_converter
+
+    def _related_image_converter(images):
+        return list(map(__class__._image_converter, images or []))
+    related_images.converter = _related_image_converter
+
+    @point_cloud.validator
+    def _point_cloud_validator(self, attribute, pcd):
+        assert pcd is None or isinstance(pcd, str), type(pcd)
 
     attributes = attrib(factory=dict, validator=default_if_none(dict))
 
@@ -494,29 +644,39 @@ class DatasetItem:
     def has_image(self):
         return self.image is not None
 
+    @property
+    def has_point_cloud(self):
+        return self.point_cloud is not None
+
     def wrap(item, **kwargs):
         return attr.evolve(item, **kwargs)
 
+
+CategoriesInfo = Dict[AnnotationType, Categories]
+
 class IExtractor:
-    def __iter__(self):
+    def __iter__(self) -> Iterable[DatasetItem]:
         raise NotImplementedError()
 
-    def __len__(self):
+    def __len__(self) -> int:
         raise NotImplementedError()
 
-    def subsets(self):
+    def __bool__(self): # avoid __len__ use for truth checking
+        return True
+
+    def subsets(self) -> Dict[str, 'IExtractor']:
         raise NotImplementedError()
 
-    def get_subset(self, name):
+    def get_subset(self, name) -> 'IExtractor':
         raise NotImplementedError()
 
-    def categories(self):
+    def categories(self) -> CategoriesInfo:
         raise NotImplementedError()
 
-    def select(self, pred):
+    def get(self, id, subset=None) -> Optional[DatasetItem]:
         raise NotImplementedError()
 
-class Extractor(IExtractor):
+class ExtractorBase(IExtractor):
     def __init__(self, length=None, subsets=None):
         self._length = length
         self._subsets = subsets
@@ -548,29 +708,51 @@ class Extractor(IExtractor):
         if self._subsets is None:
             self._init_cache()
         if name in self._subsets:
-            return self.select(lambda item: item.subset == name)
+            if len(self._subsets) == 1:
+                return self
+
+            subset = self.select(lambda item: item.subset == name)
+            subset._subsets = [name]
+            return subset
         else:
-            raise Exception("Unknown subset '%s', available subsets: %s" % \
+            raise KeyError("Unknown subset '%s', available subsets: %s" % \
                 (name, set(self._subsets)))
 
     def transform(self, method, *args, **kwargs):
         return method(self, *args, **kwargs)
 
     def select(self, pred):
-        class _DatasetFilter(Extractor):
-            def __init__(self, _):
-                super().__init__()
+        class _DatasetFilter(ExtractorBase):
             def __iter__(_):
                 return filter(pred, iter(self))
             def categories(_):
                 return self.categories()
 
-        return self.transform(_DatasetFilter)
+        return _DatasetFilter()
 
     def categories(self):
         return {}
 
+    def get(self, id, subset=None):
+        subset = subset or DEFAULT_SUBSET_NAME
+        for item in self:
+            if item.id == id and item.subset == subset:
+                return item
+        return None
+
+class Extractor(ExtractorBase):
+    """
+    A base class for user-defined and built-in extractors.
+    Should be used in cases, where SourceExtractor is not enough,
+    or its use makes problems with performance, implementation etc.
+    """
+
 class SourceExtractor(Extractor):
+    """
+    A base class for simple, single-subset extractors.
+    Should be used by default for user-defined extractors.
+    """
+
     def __init__(self, length=None, subset=None):
         self._subset = subset or DEFAULT_SUBSET_NAME
         super().__init__(length=length, subsets=[self._subset])
@@ -587,6 +769,10 @@ class SourceExtractor(Extractor):
     def __len__(self):
         return len(self._items)
 
+    def get(self, id, subset=None):
+        assert subset == self._subset, '%s != %s' % (subset, self._subset)
+        return super().get(id, subset or self._subset)
+
 class Importer:
     @classmethod
     def detect(cls, path):
@@ -597,12 +783,12 @@ class Importer:
         raise NotImplementedError()
 
     def __call__(self, path, **extra_params):
-        from datumaro.components.project import Project # cyclic import
+        from datumaro.components.project import Project  # cyclic import
         project = Project()
 
         sources = self.find_sources(osp.normpath(path))
         if len(sources) == 0:
-            raise Exception("Failed to find dataset at '%s'" % path)
+            raise DatasetNotFoundError(path)
 
         for desc in sources:
             params = dict(extra_params)
@@ -615,9 +801,39 @@ class Importer:
         return project
 
     @classmethod
-    def _find_sources_recursive(cls, path, ext, extractor_name,
-            filename='*', dirname='', file_filter=None, max_depth=3):
-        if path.endswith(ext) and osp.isfile(path):
+    def _find_sources_recursive(cls, path: str, ext: Optional[str],
+            extractor_name: str, filename: str = '*', dirname: str = '',
+            file_filter: Optional[Callable[[str], bool]] = None,
+            max_depth: int = 3):
+        """
+        Finds sources in the specified location, using the matching pattern
+        to filter file names and directories.
+        Supposed to be used, and to be the only call in subclasses.
+
+        Parameters:
+        - path - a directory or file path, where sources need to be found.
+        - ext - file extension to match. To match directories,
+            set this parameter to None or ''. Comparison is case-independent,
+            a starting dot is not required.
+        - extractor_name - the name of the associated Extractor type
+        - filename - a glob pattern for file names
+        - dirname - a glob pattern for filename prefixes
+        - file_filter - a callable (abspath: str) -> bool, to filter paths found
+        - max_depth - the maximum depth for recursive search.
+
+        Returns: a list of source configurations
+            (i.e. Extractor type names and c-tor parameters)
+        """
+
+        if ext:
+            if not ext.startswith('.'):
+                ext = '.' + ext
+            ext = ext.lower()
+
+        if (path.lower().endswith(ext) and osp.isfile(path)) or \
+                (not ext and dirname and osp.isdir(path) and \
+                os.sep + osp.normpath(dirname.lower()) + os.sep in \
+                    osp.abspath(path.lower()) + os.sep):
             sources = [{'url': path, 'format': extractor_name}]
         else:
             sources = []
@@ -630,7 +846,12 @@ class Importer:
                     break
         return sources
 
-class Transform(Extractor):
+class Transform(ExtractorBase):
+    """
+    A base class for dataset transformations that change dataset items
+    or their annotations.
+    """
+
     @staticmethod
     def wrap_item(item, **kwargs):
         return item.wrap(**kwargs)
@@ -640,10 +861,6 @@ class Transform(Extractor):
 
         self._extractor = extractor
 
-    def __iter__(self):
-        for item in self._extractor:
-            yield self.transform_item(item)
-
     def categories(self):
         return self._extractor.categories()
 
@@ -655,10 +872,24 @@ class Transform(Extractor):
     def __len__(self):
         assert self._length in {None, 'parent'} or isinstance(self._length, int)
         if self._length is None and \
-                    self.__iter__.__func__ == Transform.__iter__ \
+                    not is_method_redefined('__iter__', Transform, self) \
                 or self._length == 'parent':
             self._length = len(self._extractor)
         return super().__len__()
 
-    def transform_item(self, item: DatasetItem) -> DatasetItem:
+class ItemTransform(Transform):
+    def transform_item(self, item: DatasetItem) -> Optional[DatasetItem]:
+        """
+        Returns a modified copy of the input item.
+
+        Avoid changing and returning the input item, because it can lead to
+        unexpected problems. Use wrap_item() or item.wrap() to simplify copying.
+        """
+
         raise NotImplementedError()
+
+    def __iter__(self):
+        for item in self._extractor:
+            item = self.transform_item(item)
+            if item is not None:
+                yield item
