@@ -1,27 +1,37 @@
-
-# Copyright (C) 2020 Intel Corporation
+# Copyright (C) 2020-2021 Intel Corporation
 #
 # SPDX-License-Identifier: MIT
 
 from collections import OrderedDict
 from copy import deepcopy
+from typing import Callable, Optional
+from unittest import TestCase
 import hashlib
 import logging as log
 
+from attr import attrib, attrs
 import attr
 import cv2
 import numpy as np
-from attr import attrib, attrs
-from unittest import TestCase
 
 from datumaro.components.cli_plugin import CliPlugin
-from datumaro.components.extractor import (AnnotationType, Bbox, Label,
-    LabelCategories, PointsCategories, MaskCategories)
-from datumaro.components.project import Dataset
-from datumaro.util import find, filter_dict
-from datumaro.util.attrs_util import ensure_cls, default_if_none
-from datumaro.util.annotation_util import (segment_iou, bbox_iou,
-    mean_bbox, OKS, find_instances, max_bbox, smooth_line)
+from datumaro.components.dataset import Dataset, DatasetItemStorage, IDataset
+from datumaro.components.errors import (
+    AnnotationsTooCloseError, ConflictingCategoriesError, DatasetMergeError,
+    FailedAttrVotingError, FailedLabelVotingError, MismatchingImageInfoError,
+    NoMatchingAnnError, NoMatchingItemError, WrongGroupError,
+)
+from datumaro.components.extractor import (
+    AnnotationType, Bbox, CategoriesInfo, Label, LabelCategories,
+    MaskCategories, PointsCategories,
+)
+from datumaro.util import filter_dict, find
+from datumaro.util.annotation_util import (
+    OKS, bbox_iou, find_instances, max_bbox, mean_bbox, segment_iou,
+    smooth_line,
+)
+from datumaro.util.attrs_util import default_if_none, ensure_cls
+
 
 def get_ann_type(anns, t):
     return [a for a in anns if a.type == t]
@@ -48,13 +58,17 @@ def merge_annotations_equal(a, b):
 
 def merge_categories(sources):
     categories = {}
-    for source in sources:
+    for source_idx, source in enumerate(sources):
         for cat_type, source_cat in source.items():
             existing_cat = categories.setdefault(cat_type, source_cat)
-            if existing_cat != source_cat:
-                raise NotImplementedError(
-                    "Merging of datasets with different categories is "
-                    "only allowed in 'merge' command.")
+            if existing_cat != source_cat and len(source_cat) != 0:
+                if len(existing_cat) == 0:
+                    categories[cat_type] = source_cat
+                else:
+                    raise ConflictingCategoriesError(
+                        "Merging of datasets with different categories is "
+                        "only allowed in 'merge' command.",
+                        sources=list(range(source_idx)))
     return categories
 
 class MergingStrategy(CliPlugin):
@@ -70,75 +84,69 @@ class MergingStrategy(CliPlugin):
     def __call__(self, sources):
         raise NotImplementedError()
 
+class ExactMerge:
+    @classmethod
+    def merge(cls, *sources):
+        items = DatasetItemStorage()
+        for source_idx, source in enumerate(sources):
+            for item in source:
+                existing_item = items.get(item.id, item.subset)
+                if existing_item is not None:
+                    path = existing_item.path
+                    if item.path != path:
+                        path = None
+                    try:
+                        item = cls.merge_items(existing_item, item, path=path)
+                    except DatasetMergeError as e:
+                        e.sources = set(range(source_idx))
+                        raise e
 
-@attrs
-class DatasetError:
-    item_id = attrib()
+                items.put(item)
+        return items
 
-@attrs
-class QualityError(DatasetError):
-    pass
+    @classmethod
+    def merge_items(cls, existing_item, current_item, path=None):
+        return existing_item.wrap(path=path,
+            image=cls.merge_images(existing_item, current_item),
+            annotations=cls.merge_anno(
+                existing_item.annotations, current_item.annotations))
 
-@attrs
-class TooCloseError(QualityError):
-    a = attrib()
-    b = attrib()
-    distance = attrib()
+    @staticmethod
+    def merge_images(existing_item, current_item):
+        image = None
+        if existing_item.has_image and current_item.has_image:
+            if existing_item.image.has_data:
+                image = existing_item.image
+            else:
+                image = current_item.image
 
-    def __str__(self):
-        return "Item %s: annotations are too close: %s, %s, distance = %s" % \
-            (self.item_id, self.a, self.b, self.distance)
+            if existing_item.image.path != current_item.image.path:
+                if not existing_item.image.path:
+                    image._path = current_item.image.path
 
-@attrs
-class WrongGroupError(QualityError):
-    found = attrib(converter=set)
-    expected = attrib(converter=set)
-    group = attrib(converter=list)
+            if all([existing_item.image._size, current_item.image._size]):
+                if existing_item.image._size != current_item.image._size:
+                    raise MismatchingImageInfoError(
+                        (existing_item.id, existing_item.subset),
+                        existing_item.image._size, current_item.image._size)
+            elif existing_item.image._size:
+                image._size = existing_item.image._size
+            else:
+                image._size = current_item.image._size
+        elif existing_item.has_image:
+            image = existing_item.image
+        else:
+            image = current_item.image
 
-    def __str__(self):
-        return "Item %s: annotation group has wrong labels: " \
-            "found %s, expected %s, group %s" % \
-            (self.item_id, self.found, self.expected, self.group)
+        return image
 
-@attrs
-class MergeError(DatasetError):
-    sources = attrib(converter=set)
+    @staticmethod
+    def merge_anno(a, b):
+        return merge_annotations_equal(a, b)
 
-@attrs
-class NoMatchingAnnError(MergeError):
-    ann = attrib()
-
-    def __str__(self):
-        return "Item %s: can't find matching annotation " \
-            "in sources %s, annotation is %s" % \
-            (self.item_id, self.sources, self.ann)
-
-@attrs
-class NoMatchingItemError(MergeError):
-    def __str__(self):
-        return "Item %s: can't find matching item in sources %s" % \
-            (self.item_id, self.sources)
-
-@attrs
-class FailedLabelVotingError(MergeError):
-    votes = attrib()
-    ann = attrib(default=None)
-
-    def __str__(self):
-        return "Item %s: label voting failed%s, votes %s, sources %s" % \
-            (self.item_id, 'for ann %s' % self.ann if self.ann else '',
-            self.votes, self.sources)
-
-@attrs
-class FailedAttrVotingError(MergeError):
-    attr = attrib()
-    votes = attrib()
-    ann = attrib()
-
-    def __str__(self):
-        return "Item %s: attribute voting failed " \
-            "for ann %s, votes %s, sources %s" % \
-            (self.item_id, self.ann, self.votes, self.sources)
+    @staticmethod
+    def merge_categories(sources):
+        return merge_categories(sources)
 
 @attrs
 class IntersectMerge(MergingStrategy):
@@ -151,7 +159,7 @@ class IntersectMerge(MergingStrategy):
         quorum = attrib(converter=int, default=0)
         ignored_attributes = attrib(converter=set, factory=set)
 
-        def _groups_conveter(value):
+        def _groups_converter(value):
             result = []
             for group in value:
                 rg = set()
@@ -161,7 +169,7 @@ class IntersectMerge(MergingStrategy):
                     rg.add((name, optional))
                 result.append(rg)
             return result
-        groups = attrib(converter=_groups_conveter, factory=list)
+        groups = attrib(converter=_groups_converter, factory=list)
         close_distance = attrib(converter=float, default=0.75)
     conf = attrib(converter=ensure_cls(Conf), factory=Conf)
 
@@ -198,7 +206,8 @@ class IntersectMerge(MergingStrategy):
                 missing_sources = set(id(s) for s in datasets) - set(items)
                 missing_sources = [self._dataset_map[s][1]
                     for s in missing_sources]
-                self.add_item_error(NoMatchingItemError, missing_sources)
+                self.add_item_error(NoMatchingItemError,
+                    sources=missing_sources)
             merged.put(self.merge_items(items))
 
         return merged
@@ -275,12 +284,10 @@ class IntersectMerge(MergingStrategy):
         for (item_id, item_subset) in sorted(item_ids, key=lambda e: e[0]):
             items = {}
             for d in datasets:
-                try:
-                    item = d.get(item_id, subset=item_subset)
+                item = d.get(item_id, subset=item_subset)
+                if item:
                     items[id(d)] = item
                     item_map[id(item)] = (item, id(d))
-                except KeyError:
-                    pass
             matches[(item_id, item_subset)] = items
 
         return matches, item_map
@@ -311,11 +318,12 @@ class IntersectMerge(MergingStrategy):
                     if dst_label != src_label:
                         if src_label.parent and dst_label.parent and \
                                 src_label.parent != dst_label.parent:
-                            raise ValueError("Can't merge label category "
-                                "%s (from #%s): "
-                                "parent label conflict: %s vs. %s" % \
-                                (src_label.name, src_id,
-                                 src_label.parent, dst_label.parent)
+                            raise ConflictingCategoriesError(
+                                "Can't merge label category %s (from #%s): "
+                                    "parent label conflict: %s vs. %s" % \
+                                    (src_label.name, src_id,
+                                        src_label.parent, dst_label.parent),
+                                sources=list(range(src_id))
                             )
                         dst_label.parent = dst_label.parent or src_label.parent
                         dst_label.attributes |= src_label.attributes
@@ -342,9 +350,11 @@ class IntersectMerge(MergingStrategy):
                 dst_cat = dst_point_cat.items.get(dst_label_id)
                 if dst_cat is not None:
                     if dst_cat != src_cat:
-                        raise ValueError("Can't merge point category for label "
-                            "%s (from #%s): %s vs. %s" % \
-                            (src_label, src_id, src_cat, dst_cat)
+                        raise ConflictingCategoriesError(
+                            "Can't merge point category for label "
+                                "%s (from #%s): %s vs. %s" % \
+                                (src_label, src_id, src_cat, dst_cat),
+                            sources=list(range(src_id))
                         )
                     else:
                         pass
@@ -372,9 +382,11 @@ class IntersectMerge(MergingStrategy):
                 dst_cat = dst_mask_cat.colormap.get(dst_label_id)
                 if dst_cat is not None:
                     if dst_cat != src_cat:
-                        raise ValueError("Can't merge mask category for label "
-                            "%s (from #%s): %s vs. %s" % \
-                            (src_label, src_id, src_cat, dst_cat)
+                        raise ConflictingCategoriesError(
+                            "Can't merge mask category for label "
+                                "%s (from #%s): %s vs. %s" % \
+                                (src_label, src_id, src_cat, dst_cat),
+                            sources=list(range(src_id))
                         )
                     else:
                         pass
@@ -442,6 +454,8 @@ class IntersectMerge(MergingStrategy):
                 return _make(PointsMerger, **kwargs)
             elif t is AnnotationType.caption:
                 return _make(CaptionsMerger, **kwargs)
+            elif t is AnnotationType.cuboid_3d:
+                return _make(Cuboid3dMerger, **kwargs)
             else:
                 raise NotImplementedError("Type %s is not supported" % t)
 
@@ -527,7 +541,7 @@ class IntersectMerge(MergingStrategy):
                 missing_sources = [self._dataset_map[s][1]
                     for s in missing_sources]
                 self.add_item_error(FailedAttrVotingError,
-                    missing_sources, name, votes, ann)
+                    name, votes, ann, sources=missing_sources)
                 continue
             attributes[name] = winner
 
@@ -538,27 +552,28 @@ class IntersectMerge(MergingStrategy):
             return
 
         def _has_item(s):
-            try:
-                item =self._dataset_map[s][0].get(*self._item_id)
-                if len(item.annotations) == 0:
-                    return False
-                return True
-            except KeyError:
+            item = self._dataset_map[s][0].get(*self._item_id)
+            if not item:
                 return False
+            if len(item.annotations) == 0:
+                return False
+            return True
 
         missing_sources = set(self._dataset_map) - \
             set(self.get_ann_source(id(a)) for a in cluster)
         missing_sources = [self._dataset_map[s][1] for s in missing_sources
             if _has_item(s)]
         if missing_sources:
-            self.add_item_error(NoMatchingAnnError, missing_sources, cluster[0])
+            self.add_item_error(NoMatchingAnnError, cluster[0],
+                sources=missing_sources)
 
     def _check_annotation_distance(self, t, annotations):
         for a_idx, a_ann in enumerate(annotations):
             for b_ann in annotations[a_idx+1:]:
                 d = self._mergers[t].distance(a_ann, b_ann)
                 if self.conf.close_distance < d:
-                    self.add_item_error(TooCloseError, a_ann, b_ann, d)
+                    self.add_item_error(
+                        AnnotationsTooCloseError, a_ann, b_ann, d)
 
     def _check_groups(self, annotations):
         check_groups = []
@@ -776,6 +791,11 @@ class CaptionsMatcher(AnnotationMatcher):
     def match_annotations(self, sources):
         raise NotImplementedError()
 
+@attrs
+class Cuboid3dMatcher(_ShapeMatcher):
+    @staticmethod
+    def distance(a, b):
+        raise NotImplementedError()
 
 @attrs(kw_only=True)
 class AnnotationMerger:
@@ -803,8 +823,8 @@ class LabelMerger(AnnotationMerger, LabelMatcher):
                     if label not in [self._context._get_src_label_name(l, l.label)
                         for l in a])
                 sources = [self._context._dataset_map[s][1] for s in sources]
-                self._context.add_item_error(FailedLabelVotingError,
-                    sources, votes)
+                self._context.add_item_error(FailedLabelVotingError, votes,
+                    sources=sources)
                 continue
 
             merged.append(Label(self._context._get_label_id(label), attributes={
@@ -818,19 +838,7 @@ class _ShapeMerger(AnnotationMerger, _ShapeMatcher):
     quorum = attrib(converter=int, default=0)
 
     def merge_clusters(self, clusters):
-        merged = []
-        for cluster in clusters:
-            label, label_score = self.find_cluster_label(cluster)
-            shape, shape_score = self.merge_cluster_shape(cluster)
-
-            shape.z_order = max(cluster, key=lambda a: a.z_order).z_order
-            shape.label = label
-            shape.attributes['score'] = label_score * shape_score \
-                if label is not None else shape_score
-
-            merged.append(shape)
-
-        return merged
+        return list(map(self.merge_cluster, clusters))
 
     def find_cluster_label(self, cluster):
         votes = {}
@@ -861,6 +869,17 @@ class _ShapeMerger(AnnotationMerger, _ShapeMatcher):
             for s in cluster) / len(cluster)
         return shape, shape_score
 
+    def merge_cluster(self, cluster):
+        label, label_score = self.find_cluster_label(cluster)
+        shape, shape_score = self.merge_cluster_shape(cluster)
+
+        shape.z_order = max(cluster, key=lambda a: a.z_order).z_order
+        shape.label = label
+        shape.attributes['score'] = label_score * shape_score \
+            if label is not None else shape_score
+
+        return shape
+
 @attrs
 class BboxMerger(_ShapeMerger, BboxMatcher):
     pass
@@ -885,6 +904,26 @@ class LineMerger(_ShapeMerger, LineMatcher):
 class CaptionsMerger(AnnotationMerger, CaptionsMatcher):
     pass
 
+@attrs
+class Cuboid3dMerger(_ShapeMerger, Cuboid3dMatcher):
+    @staticmethod
+    def _merge_cluster_shape_mean_box_nearest(cluster):
+        raise NotImplementedError()
+        # mbbox = Bbox(*mean_cuboid(cluster))
+        # dist = (segment_iou(mbbox, s) for s in cluster)
+        # nearest_pos, _ = max(enumerate(dist), key=lambda e: e[1])
+        # return cluster[nearest_pos]
+
+    def merge_cluster(self, cluster):
+        label, label_score = self.find_cluster_label(cluster)
+        shape, shape_score = self.merge_cluster_shape(cluster)
+
+        shape.label = label
+        shape.attributes['score'] = label_score * shape_score \
+            if label is not None else shape_score
+
+        return shape
+
 def match_segments(a_segms, b_segms, distance=segment_iou, dist_thresh=1.0,
         label_matcher=lambda a, b: a.label == b.label):
     assert callable(distance), distance
@@ -957,6 +996,11 @@ def mean_std(dataset):
     var = lambda i, s: s[i][1]
 
     for i, item in enumerate(dataset):
+        size = item.image.size
+        if size is None:
+            log.warning("Item %s: can't detect image size, "
+                "the image will be skipped from pixel statistics", item.id)
+            continue
         counts[i] = np.prod(item.image.size)
 
         image = item.image.data
@@ -1027,8 +1071,8 @@ def compute_image_statistics(dataset):
         for item in extractor:
             if not (item.has_image and item.image.has_data):
                 available = False
-                log.warn("Item %s has no image. Image stats won't be computed",
-                    item.id)
+                log.warning("Item %s has no image, it will be excluded from "
+                    "image stats", item.id)
                 break
 
         stats = {
@@ -1059,7 +1103,7 @@ def compute_image_statistics(dataset):
     return stats
 
 def compute_ann_statistics(dataset):
-    labels = dataset.categories().get(AnnotationType.label)
+    labels = dataset.categories().get(AnnotationType.label, LabelCategories())
     def get_label(ann):
         return labels.items[ann.label].name if ann.label is not None else None
 
@@ -1174,31 +1218,10 @@ def compute_ann_statistics(dataset):
 class DistanceComparator:
     iou_threshold = attrib(converter=float, default=0.5)
 
-    @staticmethod
-    def match_datasets(a, b):
-        a_items = set((item.id, item.subset) for item in a)
-        b_items = set((item.id, item.subset) for item in b)
-
-        matches = a_items & b_items
-        a_unmatched = a_items - b_items
-        b_unmatched = b_items - a_items
-        return matches, a_unmatched, b_unmatched
-
-    @staticmethod
-    def match_classes(a, b):
-        a_label_cat = a.categories().get(AnnotationType.label, LabelCategories())
-        b_label_cat = b.categories().get(AnnotationType.label, LabelCategories())
-
-        a_labels = set(c.name for c in a_label_cat)
-        b_labels = set(c.name for c in b_label_cat)
-
-        matches = a_labels & b_labels
-        a_unmatched = a_labels - b_labels
-        b_unmatched = b_labels - a_labels
-        return matches, a_unmatched, b_unmatched
-
     def match_annotations(self, item_a, item_b):
-        return { t: self._match_ann_type(t, item_a, item_b) }
+        return { t: self._match_ann_type(t, item_a, item_b)
+            for t in AnnotationType
+        }
 
     def _match_ann_type(self, t, *args):
         # pylint: disable=no-value-for-parameter
@@ -1272,7 +1295,7 @@ class DistanceComparator:
         return match_segments(a_lines, b_lines,
             dist_thresh=self.iou_threshold, distance=matcher.distance)
 
-def match_items_by_id(a, b):
+def match_items_by_id(a: IDataset, b: IDataset):
     a_items = set((item.id, item.subset) for item in a)
     b_items = set((item.id, item.subset) for item in b)
 
@@ -1282,7 +1305,7 @@ def match_items_by_id(a, b):
     b_unmatched = b_items - a_items
     return matches, a_unmatched, b_unmatched
 
-def match_items_by_image_hash(a, b):
+def match_items_by_image_hash(a: IDataset, b: IDataset):
     a_hash = find_unique_images(a)
     b_hash = find_unique_images(b)
 
@@ -1299,7 +1322,7 @@ def match_items_by_image_hash(a, b):
 
     return matches, a_unmatched, b_unmatched
 
-def find_unique_images(dataset, item_hash=None):
+def find_unique_images(dataset: IDataset, item_hash: Optional[Callable] = None):
     def _default_hash(item):
         if not item.image or not item.image.has_data:
             if item.image and item.image.path:
@@ -1308,7 +1331,8 @@ def find_unique_images(dataset, item_hash=None):
             log.warning("Item (%s, %s) has no image "
                 "info, counted as unique", item.id, item.subset)
             return None
-        return hashlib.md5(item.image.data.tobytes()).hexdigest()
+        # ignore B303 (md5 check), because the hash is not used in a security context
+        return hashlib.md5(item.image.data.tobytes()).hexdigest() # nosec
 
     if item_hash is None:
         item_hash = _default_hash
@@ -1321,6 +1345,18 @@ def find_unique_images(dataset, item_hash=None):
         unique.setdefault(h, set()).add((item.id, item.subset))
     return unique
 
+def match_classes(a: CategoriesInfo, b: CategoriesInfo):
+    a_label_cat = a.get(AnnotationType.label, LabelCategories())
+    b_label_cat = b.get(AnnotationType.label, LabelCategories())
+
+    a_labels = set(c.name for c in a_label_cat)
+    b_labels = set(c.name for c in b_label_cat)
+
+    matches = a_labels & b_labels
+    a_unmatched = a_labels - b_labels
+    b_unmatched = b_labels - a_labels
+    return matches, a_unmatched, b_unmatched
+
 @attrs
 class ExactComparator:
     match_images = attrib(kw_only=True, type=bool, default=False)
@@ -1512,4 +1548,4 @@ class ExactComparator:
             a_unmatched |= set(a_id for a_id, m in a_matches.items() if not m)
             b_unmatched |= set(b_id for b_id, m in b_matches.items() if not m)
 
-        return matched, unmatched, a_unmatched, b_unmatched, errors
\ No newline at end of file
+        return matched, unmatched, a_unmatched, b_unmatched, errors
