@@ -1,27 +1,33 @@
-
-# Copyright (C) 2020 Intel Corporation
+# Copyright (C) 2020-2021 Intel Corporation
 #
 # SPDX-License-Identifier: MIT
 
+from enum import Enum, auto
+from itertools import chain, groupby
 import json
 import logging as log
 import os
 import os.path as osp
-from enum import Enum
-from itertools import groupby
 
 import pycocotools.mask as mask_utils
 
+from datumaro.components.converter import Converter
+from datumaro.components.dataset import ItemStatus
+from datumaro.components.extractor import (
+    _COORDINATE_ROUNDING_DIGITS, AnnotationType, DatasetItem, Points,
+)
+from datumaro.util import cast, find, str_to_bool
+from datumaro.util.image import save_image
 import datumaro.util.annotation_util as anno_tools
 import datumaro.util.mask_tools as mask_tools
-from datumaro.components.converter import Converter
-from datumaro.components.extractor import (_COORDINATE_ROUNDING_DIGITS,
-    AnnotationType, Points)
-from datumaro.util import cast, find, str_to_bool
 
 from .format import CocoPath, CocoTask
 
-SegmentationMode = Enum('SegmentationMode', ['guess', 'polygons', 'mask'])
+
+class SegmentationMode(Enum):
+    guess = auto()
+    polygons = auto()
+    mask = auto()
 
 class _TaskConverter:
     def __init__(self, context):
@@ -59,11 +65,10 @@ class _TaskConverter:
         return self._context._get_image_id(item)
 
     def save_image_info(self, item, filename):
-        if item.has_image:
+        h = 0
+        w = 0
+        if item.has_image and item.image.size:
             h, w = item.image.size
-        else:
-            h = 0
-            w = 0
 
         self._data['images'].append({
             'id': self._get_image_id(item),
@@ -89,8 +94,8 @@ class _TaskConverter:
                 ann['id'] = next_id
                 next_id += 1
 
-        with open(path, 'w') as outfile:
-            json.dump(self._data, outfile)
+        with open(path, 'w', encoding='utf-8') as outfile:
+            json.dump(self._data, outfile, ensure_ascii=False)
 
     @property
     def annotations(self):
@@ -224,16 +229,17 @@ class _InstancesConverter(_TaskConverter):
                 mask = mask_tools.rles_to_mask(polygons, img_width, img_height)
 
             if masks:
+                masks = (m.image for m in masks)
                 if mask is not None:
-                    masks += [mask]
-                mask = mask_tools.merge_masks([m.image for m in masks])
+                    masks = chain(masks, [mask])
+                mask = mask_tools.merge_masks(masks)
 
             if mask is not None:
                 mask = mask_tools.mask_to_rle(mask)
             polygons = []
         else:
             if masks:
-                mask = mask_tools.merge_masks([m.image for m in masks])
+                mask = mask_tools.merge_masks(m.image for m in masks)
                 polygons += mask_tools.mask_to_polygons(mask)
             mask = None
 
@@ -255,8 +261,8 @@ class _InstancesConverter(_TaskConverter):
         if not instances:
             return
 
-        if not item.has_image:
-            log.warn("Item '%s': skipping writing instances "
+        if not item.has_image or not item.image.size:
+            log.warning("Item '%s': skipping writing instances "
                 "since no image info available" % item.id)
             return
         h, w = item.image.size
@@ -284,10 +290,10 @@ class _InstancesConverter(_TaskConverter):
 
         area = 0
         if segmentation:
-            if item.has_image:
+            if item.has_image and item.image.size:
                 h, w = item.image.size
             else:
-                # NOTE: here we can guess the image size as
+                # Here we can guess the image size as
                 # it is only needed for the area computation
                 w = bbox[0] + bbox[2]
                 h = bbox[1] + bbox[3]
@@ -444,6 +450,67 @@ class _LabelsConverter(_TaskConverter):
 
             self.annotations.append(elem)
 
+class _StuffConverter(_InstancesConverter):
+    pass
+
+class _PanopticConverter(_TaskConverter):
+    def write(self, path):
+        with open(path, 'w', encoding='utf-8') as outfile:
+            json.dump(self._data, outfile, ensure_ascii=False)
+
+    def save_categories(self, dataset):
+        label_categories = dataset.categories().get(AnnotationType.label)
+        if label_categories is None:
+            return
+
+        for idx, cat in enumerate(label_categories.items):
+            self.categories.append({
+                'id': 1 + idx,
+                'name': cast(cat.name, str, ''),
+                'supercategory': cast(cat.parent, str, ''),
+                'isthing': 0, # TODO: can't represent this information yet
+            })
+
+    def save_annotations(self, item):
+        if not item.has_image:
+            return
+
+        ann_filename = item.id + CocoPath.PANOPTIC_EXT
+
+        segments_info = list()
+        masks = []
+        next_id = self._min_ann_id
+        for ann in item.annotations:
+            if ann.type != AnnotationType.mask:
+                continue
+
+            if not ann.id:
+                ann.id = next_id
+                next_id += 1
+
+            segment_info = {}
+            segment_info['id'] = ann.id
+            segment_info['category_id'] = cast(ann.label, int, -1) + 1
+            segment_info['area'] = float(ann.get_area())
+            segment_info['bbox'] = [float(p) for p in ann.get_bbox()]
+            segment_info['iscrowd'] = cast(ann.attributes.get("is_crowd"), int, 0)
+            segments_info.append(segment_info)
+            masks.append(ann)
+
+        if not masks:
+            return
+
+        pan_format = mask_tools.merge_masks((m.image, m.id) for m in masks)
+        save_image(osp.join(self._context._segmentation_dir, ann_filename),
+            mask_tools.index2bgr(pan_format), create_dir=True)
+
+        elem = {
+            'image_id': self._get_image_id(item),
+            'file_name': ann_filename,
+            'segments_info': segments_info
+        }
+        self.annotations.append(elem)
+
 class CocoConverter(Converter):
     @staticmethod
     def _split_tasks_string(s):
@@ -471,9 +538,12 @@ class CocoConverter(Converter):
         parser.add_argument('--allow-attributes',
             type=str_to_bool, default=True,
             help="Allow export of attributes (default: %(default)s)")
-        parser.add_argument('--reindex', action='store_true',
-            help="Assign new indices to images and annotations "
-                "(default: %(default)s)")
+        parser.add_argument('--reindex', type=str_to_bool, default=False,
+            help="Assign new indices to images and annotations, "
+                "useful to avoid merge conflicts (default: %(default)s)")
+        parser.add_argument('--merge-images', type=str_to_bool, default=False,
+            help="Save all images into a single "
+                "directory (default: %(default)s)")
         parser.add_argument('--tasks', type=cls._split_tasks_string,
             help="COCO task filter, comma-separated list of {%s} "
                 "(default: all)" % ', '.join(t.name for t in CocoTask))
@@ -487,11 +557,14 @@ class CocoConverter(Converter):
         CocoTask.person_keypoints: _KeypointsConverter,
         CocoTask.captions: _CaptionsConverter,
         CocoTask.labels: _LabelsConverter,
+        CocoTask.panoptic: _PanopticConverter,
+        CocoTask.stuff: _StuffConverter,
     }
 
     def __init__(self, extractor, save_dir,
             tasks=None, segmentation_mode=None, crop_covered=False,
-            allow_attributes=True, reindex=False, **kwargs):
+            allow_attributes=True, reindex=False, merge_images=False,
+            **kwargs):
         super().__init__(extractor, save_dir, **kwargs)
 
         assert tasks is None or isinstance(tasks, (CocoTask, list, str))
@@ -505,6 +578,8 @@ class CocoConverter(Converter):
                     tasks[i] = CocoTask[t]
                 else:
                     assert t in CocoTask, t
+        else:
+            tasks = set()
         self._tasks = tasks
 
         assert segmentation_mode is None or \
@@ -519,9 +594,12 @@ class CocoConverter(Converter):
         self._crop_covered = crop_covered
         self._allow_attributes = allow_attributes
         self._reindex = reindex
+        self._merge_images = merge_images
 
         self._image_ids = {}
 
+        self._patch = None
+
     def _make_dirs(self):
         self._images_dir = osp.join(self._save_dir, CocoPath.IMAGES_DIR)
         os.makedirs(self._images_dir, exist_ok=True)
@@ -529,6 +607,11 @@ class CocoConverter(Converter):
         self._ann_dir = osp.join(self._save_dir, CocoPath.ANNOTATIONS_DIR)
         os.makedirs(self._ann_dir, exist_ok=True)
 
+    def _make_segmentation_dir(self, subset_name):
+        self._segmentation_dir = osp.join(self._save_dir,
+            CocoPath.ANNOTATIONS_DIR, 'panoptic_'+ subset_name)
+        os.makedirs(self._segmentation_dir, exist_ok=True)
+
     def _make_task_converter(self, task):
         if task not in self._TASK_CONVERTER:
             raise NotImplementedError()
@@ -549,10 +632,6 @@ class CocoConverter(Converter):
             self._image_ids[item.id] = image_id
         return image_id
 
-    def _save_image(self, item, path=None):
-        super()._save_image(item,
-            osp.join(self._images_dir, self._make_image_filename(item)))
-
     def apply(self):
         self._make_dirs()
 
@@ -560,10 +639,14 @@ class CocoConverter(Converter):
             task_converters = self._make_task_converters()
             for task_conv in task_converters.values():
                 task_conv.save_categories(subset)
+            if CocoTask.panoptic in task_converters:
+                self._make_segmentation_dir(subset_name)
+
             for item in subset:
                 if self._save_images:
                     if item.has_image:
-                        self._save_image(item)
+                        self._save_image(item, subdir=osp.join(self._images_dir,
+                            '' if self._merge_images else subset_name))
                     else:
                         log.debug("Item '%s' has no image info", item.id)
                 for task_conv in task_converters.values():
@@ -572,10 +655,49 @@ class CocoConverter(Converter):
                     task_conv.save_annotations(item)
 
             for task, task_conv in task_converters.items():
-                if task_conv.is_empty() and not self._tasks:
+                ann_file = osp.join(self._ann_dir,
+                    '%s_%s.json' % (task.name, subset_name))
+
+                if task_conv.is_empty() and (not self._tasks or self._patch):
+                    if task == CocoTask.panoptic:
+                        os.rmdir(self._segmentation_dir)
+                    if self._patch:
+                        if osp.isfile(ann_file):
+                            # Remove subsets that became empty
+                            os.remove(ann_file)
                     continue
-                task_conv.write(osp.join(self._ann_dir,
-                    '%s_%s.json' % (task.name, subset_name)))
+
+                task_conv.write(ann_file)
+
+    @classmethod
+    def patch(cls, dataset, patch, save_dir, **kwargs):
+        for subset in patch.updated_subsets:
+            conv = cls(dataset.get_subset(subset), save_dir=save_dir, **kwargs)
+            conv._patch = patch
+            conv.apply()
+
+        conv = cls(dataset, save_dir=save_dir, **kwargs)
+        images_dir = osp.join(save_dir, CocoPath.IMAGES_DIR)
+        for (item_id, subset), status in patch.updated_items.items():
+            if status != ItemStatus.removed:
+                item = patch.data.get(item_id, subset)
+            else:
+                item = DatasetItem(item_id, subset=subset)
+
+            if not (status == ItemStatus.removed or not item.has_image):
+                continue
+
+            # Converter supports saving in separate dirs and common image dir
+
+            image_path = osp.join(images_dir, conv._make_image_filename(item))
+            if osp.isfile(image_path):
+                os.unlink(image_path)
+
+            image_path = osp.join(images_dir, subset,
+                conv._make_image_filename(item))
+            if osp.isfile(image_path):
+                os.unlink(image_path)
+
 
 class CocoInstancesConverter(CocoConverter):
     def __init__(self, *args, **kwargs):
@@ -601,3 +723,14 @@ class CocoLabelsConverter(CocoConverter):
     def __init__(self, *args, **kwargs):
         kwargs['tasks'] = CocoTask.labels
         super().__init__(*args, **kwargs)
+
+class CocoPanopticConverter(CocoConverter):
+    def __init__(self, *args, **kwargs):
+        kwargs['tasks'] = CocoTask.panoptic
+        super().__init__(*args, **kwargs)
+
+class CocoStuffConverter(CocoConverter):
+    def __init__(self, *args, **kwargs):
+        kwargs['tasks'] = CocoTask.stuff
+        kwargs['segmentation_mode'] = SegmentationMode.mask
+        super().__init__(*args, **kwargs)
