@@ -1,9 +1,10 @@
-# Copyright (C) 2020 Intel Corporation
+# Copyright (C) 2020-2021 Intel Corporation
 #
 # SPDX-License-Identifier: MIT
 
 from collections import Counter
-from enum import Enum
+from enum import Enum, auto
+from itertools import chain
 import logging as log
 import os.path as osp
 import random
@@ -11,16 +12,18 @@ import re
 
 import pycocotools.mask as mask_utils
 
-from datumaro.components.extractor import (Transform, AnnotationType,
-    RleMask, Polygon, Bbox, Label, DEFAULT_SUBSET_NAME,
-    LabelCategories, MaskCategories, PointsCategories
-)
 from datumaro.components.cli_plugin import CliPlugin
-import datumaro.util.mask_tools as mask_tools
+from datumaro.components.extractor import (
+    DEFAULT_SUBSET_NAME, AnnotationType, Bbox, ItemTransform, Label,
+    LabelCategories, MaskCategories, PointsCategories, Polygon, RleMask,
+    Transform,
+)
+from datumaro.util import NOTSET, parse_str_enum_value
 from datumaro.util.annotation_util import find_group_leader, find_instances
+import datumaro.util.mask_tools as mask_tools
 
 
-class CropCoveredSegments(Transform, CliPlugin):
+class CropCoveredSegments(ItemTransform, CliPlugin):
     def transform_item(self, item):
         annotations = []
         segments = []
@@ -83,7 +86,7 @@ class CropCoveredSegments(Transform, CliPlugin):
         max_gid = max(anns, default=0, key=lambda x: x.group)
         return max_gid + 1
 
-class MergeInstanceSegments(Transform, CliPlugin):
+class MergeInstanceSegments(ItemTransform, CliPlugin):
     """
     Replaces instance masks and, optionally, polygons with a single mask.
     """
@@ -129,6 +132,8 @@ class MergeInstanceSegments(Transform, CliPlugin):
         masks = [a for a in instance if a.type == AnnotationType.mask]
         if not polygons and not masks:
             return []
+        if not polygons and len(masks) == 1:
+            return masks
 
         leader = find_group_leader(polygons + masks)
         instance = []
@@ -143,9 +148,9 @@ class MergeInstanceSegments(Transform, CliPlugin):
             instance += polygons # keep unused polygons
 
         if masks:
-            masks = [m.image for m in masks]
+            masks = (m.image for m in masks)
             if mask is not None:
-                masks += [mask]
+                masks = chain(masks, [mask])
             mask = mask_tools.merge_masks(masks)
 
         if mask is None:
@@ -165,7 +170,7 @@ class MergeInstanceSegments(Transform, CliPlugin):
         return find_instances(a for a in annotations
             if a.type in {AnnotationType.polygon, AnnotationType.mask})
 
-class PolygonsToMasks(Transform, CliPlugin):
+class PolygonsToMasks(ItemTransform, CliPlugin):
     def transform_item(self, item):
         annotations = []
         for ann in item.annotations:
@@ -186,7 +191,7 @@ class PolygonsToMasks(Transform, CliPlugin):
         return RleMask(rle=rle, label=polygon.label, z_order=polygon.z_order,
             id=polygon.id, attributes=polygon.attributes, group=polygon.group)
 
-class BoxesToMasks(Transform, CliPlugin):
+class BoxesToMasks(ItemTransform, CliPlugin):
     def transform_item(self, item):
         annotations = []
         for ann in item.annotations:
@@ -207,7 +212,7 @@ class BoxesToMasks(Transform, CliPlugin):
         return RleMask(rle=rle, label=bbox.label, z_order=bbox.z_order,
             id=bbox.id, attributes=bbox.attributes, group=bbox.group)
 
-class MasksToPolygons(Transform, CliPlugin):
+class MasksToPolygons(ItemTransform, CliPlugin):
     def transform_item(self, item):
         annotations = []
         for ann in item.annotations:
@@ -234,7 +239,7 @@ class MasksToPolygons(Transform, CliPlugin):
             for p in polygons
         ]
 
-class ShapesToBoxes(Transform, CliPlugin):
+class ShapesToBoxes(ItemTransform, CliPlugin):
     def transform_item(self, item):
         annotations = []
         for ann in item.annotations:
@@ -270,7 +275,7 @@ class Reindex(Transform, CliPlugin):
         for i, item in enumerate(self._extractor):
             yield self.wrap_item(item, id=i + self._start)
 
-class MapSubsets(Transform, CliPlugin):
+class MapSubsets(ItemTransform, CliPlugin):
     @staticmethod
     def _mapping_arg(s):
         parts = s.split(':')
@@ -296,9 +301,9 @@ class MapSubsets(Transform, CliPlugin):
             mapping = dict(tuple(m) for m in mapping)
         self._mapping = mapping
 
-        if extractor._subsets:
+        if extractor.subsets():
             counts = Counter(mapping.get(s, s) or DEFAULT_SUBSET_NAME
-                for s in extractor._subsets)
+                for s in extractor.subsets())
             if all(c == 1 for c in counts.values()):
                 self._length = 'parent'
             self._subsets = set(counts)
@@ -383,7 +388,7 @@ class RandomSplit(Transform, CliPlugin):
         for i, item in enumerate(self._extractor):
             yield self.wrap_item(item, subset=self._find_split(i))
 
-class IdFromImageName(Transform, CliPlugin):
+class IdFromImageName(ItemTransform, CliPlugin):
     def transform_item(self, item):
         if item.has_image and item.image.path:
             name = osp.splitext(osp.basename(item.image.path))[0]
@@ -393,8 +398,8 @@ class IdFromImageName(Transform, CliPlugin):
                 "item has no image info" % item.id)
             return item
 
-class Rename(Transform, CliPlugin):
-    """
+class Rename(ItemTransform, CliPlugin):
+    r"""
     Renames items in the dataset. Supports regular expressions.
     The first character in the expression is a delimiter for
     the pattern and replacement parts. Replacement part can also
@@ -404,7 +409,7 @@ class Rename(Transform, CliPlugin):
     - Replace 'pattern' with 'replacement':|n
     |s|srename -e '|pattern|replacement|'|n
     - Remove 'frame_' from item ids:|n
-    |s|srename -e '|frame_(\d+)|\\1|'
+    |s|srename -e '|frame_(\d+)|\1|'
     """
 
     @classmethod
@@ -427,15 +432,32 @@ class Rename(Transform, CliPlugin):
         return self.wrap_item(item, id=self._re.sub(self._sub, item.id) \
             .format(item=item))
 
-class RemapLabels(Transform, CliPlugin):
+class RemapLabels(ItemTransform, CliPlugin):
     """
     Changes labels in the dataset.|n
+    |n
+    A label can be:|n
+    - renamed (and joined with existing) -|n
+    |s|swhen specified '--label <old_name>:<new_name>'|n
+    - deleted - when specified '--label <name>:' or default action is 'delete'|n
+    |s|sand the label is not mentioned in the list. When a label|n
+    |s|sis deleted, all the associated annotations are removed|n
+    - kept unchanged - when specified '--label <name>:<name>'|n
+    |s|sor default action is 'keep' and the label is not mentioned in the list|n
+    Annotations with no label are managed by the default action policy.|n
+    |n
     Examples:|n
+    - Remove the 'person' label (and corresponding annotations):|n
+    |s|sremap_labels -l person: --default keep|n
+    - Rename 'person' to 'pedestrian' and 'human' to 'pedestrian', join:|n
+    |s|sremap_labels -l person:pedestrian -l human:pedestrian --default keep|n
     - Rename 'person' to 'car' and 'cat' to 'dog', keep 'bus', remove others:|n
     |s|sremap_labels -l person:car -l bus:bus -l cat:dog --default delete
     """
 
-    DefaultAction = Enum('DefaultAction', ['keep', 'delete'])
+    class DefaultAction(Enum):
+        keep = auto()
+        delete = auto()
 
     @staticmethod
     def _split_arg(s):
@@ -460,9 +482,9 @@ class RemapLabels(Transform, CliPlugin):
     def __init__(self, extractor, mapping, default=None):
         super().__init__(extractor)
 
-        assert isinstance(default, (str, self.DefaultAction))
-        if isinstance(default, str):
-            default = self.DefaultAction[default]
+        default = parse_str_enum_value(default, self.DefaultAction,
+            self.DefaultAction.keep)
+        self._default_action = default
 
         assert isinstance(mapping, (dict, list))
         if isinstance(mapping, list):
@@ -479,20 +501,20 @@ class RemapLabels(Transform, CliPlugin):
             assert src_label_cat is not None
             dst_mask_cat = MaskCategories(attributes=src_mask_cat.attributes)
             dst_mask_cat.colormap = {
-                id: src_mask_cat.colormap[id]
+                id: src_mask_cat[id]
                 for id, _ in enumerate(src_label_cat.items)
-                if self._map_id(id) or id == 0
+                if id in src_mask_cat and (self._map_id(id) or id == 0)
             }
             self._categories[AnnotationType.mask] = dst_mask_cat
 
-        src_points_cat = self._extractor.categories().get(AnnotationType.points)
-        if src_points_cat is not None:
+        src_point_cat = self._extractor.categories().get(AnnotationType.points)
+        if src_point_cat is not None:
             assert src_label_cat is not None
-            dst_points_cat = PointsCategories(attributes=src_points_cat.attributes)
+            dst_points_cat = PointsCategories(attributes=src_point_cat.attributes)
             dst_points_cat.items = {
-                id: src_points_cat.items[id]
-                for id, item in enumerate(src_label_cat.items)
-                if self._map_id(id) or id == 0
+                id: src_point_cat[id]
+                for id, _ in enumerate(src_label_cat.items)
+                if id in src_point_cat and (self._map_id(id) or id == 0)
             }
             self._categories[AnnotationType.points] = dst_points_cat
 
@@ -500,10 +522,10 @@ class RemapLabels(Transform, CliPlugin):
         dst_label_cat = LabelCategories(attributes=src_label_cat.attributes)
         id_mapping = {}
         for src_index, src_label in enumerate(src_label_cat.items):
-            dst_label = label_mapping.get(src_label.name)
-            if not dst_label and default_action == self.DefaultAction.keep:
+            dst_label = label_mapping.get(src_label.name, NOTSET)
+            if dst_label is NOTSET and default_action == self.DefaultAction.keep:
                 dst_label = src_label.name # keep unspecified as is
-            if not dst_label:
+            elif not dst_label or dst_label is NOTSET:
                 continue
 
             dst_index = dst_label_cat.find(dst_label)[0]
@@ -515,7 +537,7 @@ class RemapLabels(Transform, CliPlugin):
         if log.getLogger().isEnabledFor(log.DEBUG):
             log.debug("Label mapping:")
             for src_id, src_label in enumerate(src_label_cat.items):
-                if id_mapping.get(src_id):
+                if id_mapping.get(src_id) is not None:
                     log.debug("#%s '%s' -> #%s '%s'",
                         src_id, src_label.name, id_mapping[src_id],
                         dst_label_cat.items[id_mapping[src_id]].name
@@ -532,18 +554,15 @@ class RemapLabels(Transform, CliPlugin):
     def transform_item(self, item):
         annotations = []
         for ann in item.annotations:
-            if ann.type in { AnnotationType.label, AnnotationType.mask,
-                AnnotationType.points, AnnotationType.polygon,
-                AnnotationType.polyline, AnnotationType.bbox
-            } and ann.label is not None:
+            if getattr(ann, 'label') is not None:
                 conv_label = self._map_id(ann.label)
                 if conv_label is not None:
                     annotations.append(ann.wrap(label=conv_label))
-            else:
+            elif self._default_action is self.DefaultAction.keep:
                 annotations.append(ann.wrap())
         return item.wrap(annotations=annotations)
 
-class AnnsToLabels(Transform, CliPlugin):
+class AnnsToLabels(ItemTransform, CliPlugin):
     """
     Collects all labels from annotations (of all types) and
     transforms them into a set of annotations of type Label
@@ -556,4 +575,21 @@ class AnnsToLabels(Transform, CliPlugin):
         for label in labels:
             annotations.append(Label(label=label))
 
-        return item.wrap(annotations=annotations)
\ No newline at end of file
+        return item.wrap(annotations=annotations)
+
+class BboxValuesDecrement(ItemTransform, CliPlugin):
+    """
+    Subtracts one from the coordinates of bounding boxes
+    """
+
+    def transform_item(self, item):
+        annotations = [p for p in item.annotations
+            if p.type != AnnotationType.bbox]
+        bboxes = [p for p in item.annotations
+            if p.type == AnnotationType.bbox]
+        for bbox in bboxes:
+            annotations.append(Bbox(
+                bbox.x - 1, bbox.y - 1, bbox.w, bbox.h,
+                label=bbox.label, attributes=bbox.attributes))
+
+        return item.wrap(annotations=annotations)
