@@ -1,30 +1,33 @@
-
-# Copyright (C) 2020 Intel Corporation
+# Copyright (C) 2020-2021 Intel Corporation
 #
 # SPDX-License-Identifier: MIT
 
+from collections import OrderedDict, defaultdict
+from enum import Enum, auto
+from itertools import chain
 import logging as log
 import os
 import os.path as osp
-from collections import OrderedDict, defaultdict
-from enum import Enum
-from itertools import chain
 
 from lxml import etree as ET
 
 from datumaro.components.converter import Converter
-from datumaro.components.extractor import (AnnotationType,
-    CompiledMask, LabelCategories)
+from datumaro.components.dataset import ItemStatus
+from datumaro.components.extractor import (
+    AnnotationType, CompiledMask, DatasetItem, LabelCategories,
+)
 from datumaro.util import find, str_to_bool
+from datumaro.util.annotation_util import make_label_id_mapping
 from datumaro.util.image import save_image
 from datumaro.util.mask_tools import paint_mask, remap_mask
 
-from .format import (VocTask, VocPath, VocInstColormap,
-    parse_label_map, make_voc_label_map, make_voc_categories, write_label_map
+from .format import (
+    VocInstColormap, VocPath, VocTask, make_voc_categories, make_voc_label_map,
+    parse_label_map, write_label_map,
 )
 
 
-def _convert_attr(name, attributes, type_conv, default=None, warn=True):
+def _convert_attr(name, attributes, type_conv, default=None):
     d = object()
     value = attributes.get(name, d)
     if value is d:
@@ -47,10 +50,13 @@ def _write_xml_bbox(bbox, parent_elem):
     return bbox_elem
 
 
-LabelmapType = Enum('LabelmapType', ['voc', 'source'])
+class LabelmapType(Enum):
+    voc = auto()
+    source = auto()
 
 class VocConverter(Converter):
     DEFAULT_IMAGE_EXT = VocPath.IMAGE_EXT
+    BUILTIN_ATTRS = {'difficult', 'pose', 'truncated', 'occluded' }
 
     @staticmethod
     def _split_tasks_string(s):
@@ -79,6 +85,10 @@ class VocConverter(Converter):
         parser.add_argument('--allow-attributes',
             type=str_to_bool, default=True,
             help="Allow export of attributes (default: %(default)s)")
+        parser.add_argument('--keep-empty',
+            type=str_to_bool, default=False,
+            help="Write subset lists even if they are empty "
+                "(default: %(default)s)")
         parser.add_argument('--tasks', type=cls._split_tasks_string,
             help="VOC task filter, comma-separated list of {%s} "
                 "(default: all)" % ', '.join(t.name for t in VocTask))
@@ -87,7 +97,7 @@ class VocConverter(Converter):
 
     def __init__(self, extractor, save_dir,
             tasks=None, apply_colormap=True, label_map=None,
-            allow_attributes=True, **kwargs):
+            allow_attributes=True, keep_empty=False, **kwargs):
         super().__init__(extractor, save_dir, **kwargs)
 
         assert tasks is None or isinstance(tasks, (VocTask, list, set))
@@ -101,12 +111,15 @@ class VocConverter(Converter):
 
         self._apply_colormap = apply_colormap
         self._allow_attributes = allow_attributes
+        self._keep_empty = keep_empty
 
         if label_map is None:
             label_map = LabelmapType.source.name
         assert isinstance(label_map, (str, dict)), label_map
         self._load_categories(label_map)
 
+        self._patch = None
+
     def apply(self):
         self.make_dirs()
         self.save_subsets()
@@ -181,9 +194,8 @@ class VocConverter(Converter):
                     elif a.type == AnnotationType.mask:
                         masks.append(a)
 
-                if self._tasks is None and bboxes or \
-                        self._tasks & {VocTask.detection, VocTask.person_layout,
-                            VocTask.action_classification}:
+                if self._tasks & {VocTask.detection, VocTask.person_layout,
+                        VocTask.action_classification}:
                     root_elem = ET.Element('annotation')
                     if '_' in item.id:
                         folder = item.id[ : item.id.find('_')]
@@ -197,7 +209,7 @@ class VocConverter(Converter):
                     ET.SubElement(source_elem, 'annotation').text = 'Unknown'
                     ET.SubElement(source_elem, 'image').text = 'Unknown'
 
-                    if item.has_image:
+                    if item.has_image and item.image.has_size:
                         h, w = item.image.size
                         size_elem = ET.SubElement(root_elem, 'size')
                         ET.SubElement(size_elem, 'width').text = str(w)
@@ -232,20 +244,12 @@ class VocConverter(Converter):
                             ET.SubElement(obj_elem, 'pose').text = \
                                 str(attr['pose'])
 
-                        if 'truncated' in attr:
-                            truncated = _convert_attr('truncated', attr, int, 0)
-                            ET.SubElement(obj_elem, 'truncated').text = \
-                                '%d' % truncated
-
-                        if 'difficult' in attr:
-                            difficult = _convert_attr('difficult', attr, int, 0)
-                            ET.SubElement(obj_elem, 'difficult').text = \
-                                '%d' % difficult
-
-                        if 'occluded' in attr:
-                            occluded = _convert_attr('occluded', attr, int, 0)
-                            ET.SubElement(obj_elem, 'occluded').text = \
-                                '%d' % occluded
+                        ET.SubElement(obj_elem, 'truncated').text = \
+                            '%d' % _convert_attr('truncated', attr, int, 0)
+                        ET.SubElement(obj_elem, 'occluded').text = \
+                            '%d' % _convert_attr('occluded', attr, int, 0)
+                        ET.SubElement(obj_elem, 'difficult').text = \
+                            '%d' % _convert_attr('difficult', attr, int, 0)
 
                         bbox = obj.get_bbox()
                         if bbox is not None:
@@ -276,8 +280,7 @@ class VocConverter(Converter):
                             obj_elem.append(actions_elem)
 
                         if self._allow_attributes:
-                            native_attrs = {'difficult', 'pose',
-                                'truncated', 'occluded' }
+                            native_attrs = set(self.BUILTIN_ATTRS)
                             native_attrs.update(label_actions)
 
                             attrs_elem = ET.Element('attributes')
@@ -294,13 +297,17 @@ class VocConverter(Converter):
                             VocTask.action_classification}:
                         ann_path = osp.join(self._ann_dir, item.id + '.xml')
                         os.makedirs(osp.dirname(ann_path), exist_ok=True)
-                        with open(ann_path, 'w') as f:
+                        with open(ann_path, 'w', encoding='utf-8') as f:
                             f.write(ET.tostring(root_elem,
                                 encoding='unicode', pretty_print=True))
 
                     clsdet_list[item.id] = True
-                    layout_list[item.id] = objects_with_parts
-                    action_list[item.id] = objects_with_actions
+
+                    if objects_with_parts:
+                        layout_list[item.id] = objects_with_parts
+
+                    if objects_with_actions:
+                        action_list[item.id] = objects_with_actions
 
                 for label_ann in labels:
                     label = self.get_label(label_ann.label)
@@ -312,7 +319,7 @@ class VocConverter(Converter):
 
                     clsdet_list[item.id] = True
 
-                if masks:
+                if masks and VocTask.segmentation in self._tasks:
                     compiled_mask = CompiledMask.from_instance_masks(masks,
                         instance_labels=[self._label_id_mapping(m.label)
                             for m in masks])
@@ -345,87 +352,163 @@ class VocConverter(Converter):
             if self._tasks & {VocTask.segmentation}:
                 self.save_segm_lists(subset_name, segm_list)
 
-    def save_action_lists(self, subset_name, action_list):
-        if not action_list:
-            return
+    @staticmethod
+    def _get_filtered_lines(path, patch, subset, items=None):
+        lines = {}
+        with open(path, encoding='utf-8') as f:
+            for line in f:
+                line = line.strip()
+                if not line:
+                    continue
+                line_parts = line.split(maxsplit=1)
+                if len(line_parts) < 2:
+                    line_parts.append('')
+                item, text = line_parts
+                if not patch or patch.updated_items.get((item, subset)) != \
+                        ItemStatus.removed:
+                    lines.setdefault(item, []).append(text)
+        if items is not None:
+            items.update((k, True) for k in lines)
+        return lines
 
+    def save_action_lists(self, subset_name, action_list):
         os.makedirs(self._action_subsets_dir, exist_ok=True)
 
         ann_file = osp.join(self._action_subsets_dir, subset_name + '.txt')
-        with open(ann_file, 'w') as f:
-            for item in action_list:
-                f.write('%s\n' % item)
+        items = {k: True for k in action_list}
+        if self._patch and osp.isfile(ann_file):
+            self._get_filtered_lines(ann_file, self._patch, subset_name, items)
 
-        if len(action_list) == 0:
+        if items or self._keep_empty:
+            with open(ann_file, 'w', encoding='utf-8') as f:
+                for item in items:
+                    f.write('%s\n' % item)
+        elif osp.isfile(ann_file):
+            os.remove(ann_file)
+
+        if not items and not self._patch and not self._keep_empty:
             return
 
-        all_actions = set(chain(*(self._get_actions(l)
-            for l in self._label_map)))
-        for action in all_actions:
-            ann_file = osp.join(self._action_subsets_dir,
-                '%s_%s.txt' % (action, subset_name))
-            with open(ann_file, 'w') as f:
-                for item, objs in action_list.items():
-                    if not objs:
-                        continue
-                    for obj_id, obj_actions in objs.items():
-                        presented = obj_actions[action]
-                        f.write('%s %s % d\n' % \
-                            (item, 1 + obj_id, 1 if presented else -1))
+        def _write_item(f, item, objs, action):
+            if not objs:
+                return
+            for obj_id, obj_actions in objs.items():
+                presented = obj_actions[action]
+                f.write('%s %s % d\n' % \
+                    (item, 1 + obj_id, 1 if presented else -1))
+
+        all_actions = {
+            act: osp.join(self._action_subsets_dir,
+                '%s_%s.txt' % (act, subset_name))
+            for act in chain(*(self._get_actions(l) for l in self._label_map))
+        }
+        for action, ann_file in all_actions.items():
+            if not items and not self._keep_empty:
+                if osp.isfile(ann_file):
+                    os.remove(ann_file)
+                continue
+
+            lines = {}
+            if self._patch and osp.isfile(ann_file):
+                lines = self._get_filtered_lines(ann_file, None, subset_name)
+
+            with open(ann_file, 'w', encoding='utf-8') as f:
+                for item in items:
+                    if item in action_list:
+                        _write_item(f, item, action_list[item], action)
+                    elif item in lines:
+                        f.writelines(lines[item])
 
     def save_class_lists(self, subset_name, class_lists):
-        if not class_lists:
-            return
+        def _write_item(f, item, item_labels):
+            if not item_labels:
+                return
+            item_labels = [self.get_label(l) for l in item_labels]
+            presented = label in item_labels
+            f.write('%s % d\n' % (item, 1 if presented else -1))
 
         os.makedirs(self._cls_subsets_dir, exist_ok=True)
 
         for label in self._label_map:
             ann_file = osp.join(self._cls_subsets_dir,
                 '%s_%s.txt' % (label, subset_name))
-            with open(ann_file, 'w') as f:
-                for item, item_labels in class_lists.items():
-                    if not item_labels:
-                        continue
-                    item_labels = [self.get_label(l) for l in item_labels]
-                    presented = label in item_labels
-                    f.write('%s % d\n' % (item, 1 if presented else -1))
+            items = {k: True for k in class_lists}
+            lines = {}
+            if self._patch and osp.isfile(ann_file):
+                lines = self._get_filtered_lines(ann_file, self._patch,
+                    subset_name, items)
+
+            if not items and not self._keep_empty:
+                if osp.isfile(ann_file):
+                    os.remove(ann_file)
+                continue
+
+            with open(ann_file, 'w', encoding='utf-8') as f:
+                for item in items:
+                    if item in class_lists:
+                        _write_item(f, item, class_lists[item])
+                    elif item in lines:
+                        f.writelines(lines[item])
 
     def save_clsdet_lists(self, subset_name, clsdet_list):
-        if not clsdet_list:
-            return
-
         os.makedirs(self._cls_subsets_dir, exist_ok=True)
 
         ann_file = osp.join(self._cls_subsets_dir, subset_name + '.txt')
-        with open(ann_file, 'w') as f:
-            for item in clsdet_list:
-                f.write('%s\n' % item)
+        items = {k: True for k in clsdet_list}
+        if self._patch and osp.isfile(ann_file):
+            self._get_filtered_lines(ann_file, self._patch, subset_name, items)
+
+        if items or self._keep_empty:
+            with open(ann_file, 'w', encoding='utf-8') as f:
+                for item in items:
+                    f.write('%s\n' % item)
+        elif osp.isfile(ann_file):
+            os.remove(ann_file)
 
     def save_segm_lists(self, subset_name, segm_list):
-        if not segm_list:
-            return
-
         os.makedirs(self._segm_subsets_dir, exist_ok=True)
 
         ann_file = osp.join(self._segm_subsets_dir, subset_name + '.txt')
-        with open(ann_file, 'w') as f:
-            for item in segm_list:
-                f.write('%s\n' % item)
+        items = {k: True for k in segm_list}
+        if self._patch and osp.isfile(ann_file):
+            self._get_filtered_lines(ann_file, self._patch, subset_name, items)
+
+        if items or self._keep_empty:
+            with open(ann_file, 'w', encoding='utf-8') as f:
+                for item in items:
+                    f.write('%s\n' % item)
+        elif osp.isfile(ann_file):
+            os.remove(ann_file)
 
     def save_layout_lists(self, subset_name, layout_list):
-        if not layout_list:
-            return
+        def _write_item(f, item, item_layouts):
+            if 1 < len(item.split()):
+                item = '\"' + item + '\"'
+            if item_layouts:
+                for obj_id in item_layouts:
+                    f.write('%s % d\n' % (item, 1 + obj_id))
+            else:
+                f.write('%s\n' % item)
 
         os.makedirs(self._layout_subsets_dir, exist_ok=True)
 
         ann_file = osp.join(self._layout_subsets_dir, subset_name + '.txt')
-        with open(ann_file, 'w') as f:
-            for item, item_layouts in layout_list.items():
-                if item_layouts:
-                    for obj_id in item_layouts:
-                        f.write('%s % d\n' % (item, 1 + obj_id))
-                else:
-                    f.write('%s\n' % (item))
+        items = {k: True for k in layout_list}
+        lines = {}
+        if self._patch and osp.isfile(ann_file):
+            self._get_filtered_lines(ann_file, self._patch, subset_name, items)
+
+        if not items and not self._keep_empty:
+            if osp.isfile(ann_file):
+                os.remove(ann_file)
+            return
+
+        with open(ann_file, 'w', encoding='utf-8') as f:
+            for item in items:
+                if item in layout_list:
+                    _write_item(f, item, layout_list[item])
+                elif item in lines:
+                    f.writelines(lines[item])
 
     def save_segm(self, path, mask, colormap=None):
         if self._apply_colormap:
@@ -517,22 +600,12 @@ class VocConverter(Converter):
         return label_desc[2]
 
     def _make_label_id_map(self):
-        source_labels = {
-            id: label.name for id, label in
-            enumerate(self._extractor.categories().get(
-                AnnotationType.label, LabelCategories()).items)
-        }
-        target_labels = {
-            label.name: id for id, label in
-            enumerate(self._categories[AnnotationType.label].items)
-        }
-        id_mapping = {
-            src_id: target_labels.get(src_label, 0)
-            for src_id, src_label in source_labels.items()
-        }
+        map_id, id_mapping, src_labels, dst_labels = make_label_id_mapping(
+            self._extractor.categories().get(AnnotationType.label),
+            self._categories[AnnotationType.label])
 
-        void_labels = [src_label for src_id, src_label in source_labels.items()
-            if src_label not in target_labels]
+        void_labels = [src_label for src_id, src_label in src_labels.items()
+            if src_label not in dst_labels]
         if void_labels:
             log.warning("The following labels are remapped to background: %s" %
                 ', '.join(void_labels))
@@ -543,17 +616,66 @@ class VocConverter(Converter):
                     self._categories[AnnotationType.label] \
                         .items[id_mapping[src_id]].name
                 )
-                for src_id, src_label in source_labels.items()
+                for src_id, src_label in src_labels.items()
             ])
         )
 
-        def map_id(src_id):
-            return id_mapping.get(src_id, 0)
         return map_id
 
     def _remap_mask(self, mask):
         return remap_mask(mask, self._label_id_mapping)
 
+    @classmethod
+    def patch(cls, dataset, patch, save_dir, **kwargs):
+        conv = cls(patch.as_dataset(dataset), save_dir=save_dir, **kwargs)
+        conv._patch = patch
+        conv.apply()
+
+        # Find images that needs to be removed
+        # images from different subsets are stored in the common directory
+        # Avoid situations like:
+        # (a, test): added
+        # (a, train): removed
+        # where the second line removes images from the first.
+        ids_to_remove = {}
+        for (item_id, subset), status in patch.updated_items.items():
+            if status != ItemStatus.removed:
+                item = patch.data.get(item_id, subset)
+            else:
+                item = DatasetItem(item_id, subset=subset)
+
+            if not (status == ItemStatus.removed or not item.has_image):
+                ids_to_remove[item_id] = (item, False)
+            else:
+                ids_to_remove.setdefault(item_id, (item, True))
+
+        for item, to_remove in ids_to_remove.values():
+            if not to_remove:
+                continue
+
+            if conv._tasks & {VocTask.detection,
+                    VocTask.action_classification, VocTask.person_layout}:
+                ann_path = osp.join(conv._ann_dir, item.id + '.xml')
+                if osp.isfile(ann_path):
+                    os.remove(ann_path)
+
+            image_path = osp.join(conv._images_dir,
+                conv._make_image_filename(item))
+            if osp.isfile(image_path):
+                os.unlink(image_path)
+
+            if not [a for a in item.annotations
+                    if a.type is AnnotationType.mask]:
+                path = osp.join(save_dir, VocPath.SEGMENTATION_DIR,
+                    item.id + VocPath.SEGM_EXT)
+                if osp.isfile(path):
+                    os.unlink(path)
+
+                path = osp.join(save_dir, VocPath.INSTANCES_DIR,
+                    item.id + VocPath.SEGM_EXT)
+                if osp.isfile(path):
+                    os.unlink(path)
+
 class VocClassificationConverter(VocConverter):
     def __init__(self, *args, **kwargs):
         kwargs['tasks'] = VocTask.classification
